import numpy as np
import onnx_graphsurgeon as gs

from convdog.simplifier.base_pass import BasePass
from convdog.utils.logger import logger

class FuseMultipleGatherPass(BasePass):
    """
    O1 优化：将同一源张量上多次对连续索引 (例如 0,1,2,...) 的 Gather 合并为一个 Split。
    修正要点：
      - 仅合并对同一 gs.Variable 的 Gather（identity check）
      - 如果原 Gather 是 scalar indices（会降维），则在 Split 后插入 Squeeze，使语义与 Gather 保持一致
      - 为 split 与 squeeze 的输出变量尽量填充 dtype/shape 以通过 shape inference
    """
    def process(self, graph: gs.Graph) -> bool:
        changed = False
        fusion_count = 0

        logger.debug("[O1]: 正在融合分组Gather......")
        graph.cleanup().toposort()

        # 收集所有 Gather 节点（使用快照）
        all_gathers = [n for n in graph.nodes if n.op == "Gather"]

        # 按 (data_var id, axis) 分组 —— 使用变量对象 identity 而不是仅 name
        groups = {}
        for g in all_gathers:
            if len(g.inputs) < 2:
                continue
            data = g.inputs[0]
            idx = g.inputs[1]
            if not isinstance(idx, gs.Constant):
                continue
            axis = int(g.attrs.get("axis", 0))
            key = (id(data), axis)  # 用 id(data) 保证同一对象
            groups.setdefault(key, []).append(g)

        for (data_id, axis), gnodes in groups.items():
            if len(gnodes) < 2:
                continue

            # 收集 (index_value, gather_node) 并检测 indices 是否为 scalar
            entries = []
            skip_group = False
            any_non_scalar = False
            for g in gnodes:
                idx_const = g.inputs[1]
                try:
                    vals = np.array(idx_const.values).astype(int).reshape(-1)
                except Exception:
                    skip_group = True
                    break
                if vals.size != 1:
                    any_non_scalar = True
                entries.append((int(vals[0]) if vals.size == 1 else tuple(vals.tolist()), g))
            if skip_group:
                continue

            # 只处理所有 indices 都是 scalar 的情况（更安全）
            if any_non_scalar:
                continue

            indices = sorted([e[0] for e in entries])
            # 暂定策略：只处理从0开始的连续序列 0..k-1
            if indices[0] != 0 or indices != list(range(indices[-1] + 1)):
                continue

            k = indices[-1] + 1

            # 源张量（取第一个 gather 的输入）
            data_var = gnodes[0].inputs[0]

            # 如果有静态 shape，则做额外校验并归一化 axis（支持负轴）
            axis_norm = axis
            if data_var.shape:
                rank = len(data_var.shape)
                if axis < 0:
                    axis_norm = axis + rank
                if axis_norm < 0 or axis_norm >= rank:
                    logger.debug(f"[O1] 跳过融合 Split: axis {axis} 越界 for {data_var.name}")
                    continue
                dim_size = data_var.shape[axis_norm]
                if dim_size is not None and dim_size < k:
                    logger.debug(f"[O1] 跳过融合 Split: 源张量 {data_var.name} 在 axis {axis_norm} 的长度 {dim_size} 小于 k={k}")
                    continue
            else:
                # 若没有 shape 信息，仍使用 axis 原值作为 split axis
                axis_norm = axis

            # 判断原始 Gather 是否会降维：取任一 gnode 的输出 shape（若有）和 data_var.shape 比较
            gather_reduces_dim = False
            sample_g_out = None
            if gnodes[0].outputs and gnodes[0].outputs[0].shape and data_var.shape:
                sample_g_out = gnodes[0].outputs[0].shape
                if len(sample_g_out) == len(data_var.shape) - 1:
                    gather_reduces_dim = True

            # 构造 split 输出 Variable（先为 split 的原始输出 —— 未 squeeze）
            split_outs = []
            for i in range(k):
                out_shape = None
                if data_var.shape:
                    out_shape = list(data_var.shape)
                    try:
                        out_shape[axis_norm] = 1  # Split 每块把该维设为 1
                    except Exception:
                        out_shape = None
                out_dtype = getattr(data_var, "dtype", None)
                split_var = gs.Variable(name=f"{data_var.name}_split_{i}", dtype=out_dtype, shape=out_shape)
                split_outs.append(split_var)

            # 创建 Split 节点，带 split & num_outputs
            split_var = gs.Constant(
                name=f"{data_var.name}_split",
                values=np.array([1] * k, dtype=np.int64)
            )
            split_node = gs.Node(
                op="Split",
                name=f"Fused_Split_{data_var.name}",
                inputs=[data_var, split_var],
                outputs=split_outs,
                attrs={"axis": int(axis_norm)}
            )
            graph.nodes.append(split_node)

            # 如果原 Gather 会降维，则为每个 split 输出插入 Squeeze 节点并将消费者指向 squeeze 输出
            squeeze_outs = {}
            if gather_reduces_dim:
                for i, split_var in enumerate(split_outs):
                    squeeze_var = gs.Variable(
                        name=f"{split_var.name}_sq",
                        dtype=split_var.dtype
                    )

                    # 修正点：将 axes 作为 Constant 输入，而不是 Attribute
                    # 这能确保在 Opset 13+ 中 Squeeze 只针对 axis_norm 操作
                    axes_const = gs.Constant(
                        name=f"{split_var.name}_axes",
                        values=np.array([axis_norm], dtype=np.int64)
                    )

                    squeeze_node = gs.Node(
                        op="Squeeze",
                        name=f"Fused_Squeeze_{data_var.name}_{i}",
                        inputs=[split_var, axes_const], # 两个输入
                        outputs=[squeeze_var]
                    )
                    graph.nodes.append(squeeze_node)
                    squeeze_outs[i] = squeeze_var

            # 将每个 Gather 的输出消费者重定向到对应 split_out（或 squeeze_out）
            for idx_val, gnode in entries:
                if not gnode.outputs:
                    continue
                out_var = gnode.outputs[0]
                target_var = squeeze_outs.get(idx_val, split_outs[idx_val])

                # 替换 downstream 的输入项
                for consumer in list(out_var.outputs):
                    for i, inp in enumerate(consumer.inputs):
                        if inp == out_var:
                            consumer.inputs[i] = target_var
                            logger.debug(f"[O1] 重定向: {consumer.name} 的输入 {out_var.name} -> {target_var.name}")

                # 记录日志并断开旧节点引用，等待 cleanup 删除
                logger.debug(f"[O1] 合并 Gather -> {split_node.name} 映射: {gnode.name} -> {target_var.name}")
                gnode.inputs = []
                gnode.outputs = []

            fusion_count += 1
            changed = True

        if changed:
            graph.cleanup().toposort()
            logger.debug(f"[O1] 总共执行了 {fusion_count} 次 Gather->Split(+Squeeze) 融合")

        return changed



import onnx_graphsurgeon as gs
from convdog.simplifier.base_pass import BasePass
from convdog.utils.logger import logger

class FuseAttentionPass(BasePass):
    Custom_Op_Map = {"Attention": "com.microsoft:Attention"}

    def process(self, graph: "gs.Graph") -> bool:
        changed = False
        graph.cleanup().toposort()

        for softmax_node in [n for n in graph.nodes if n.op == "Softmax"]:
            try:
                # --- 1. 核心算子定位 (同前) ---
                score_matmul = softmax_node.inputs[0].inputs[0] # Q * K
                context_matmul = softmax_node.outputs[0].outputs[0] # Prob * V
                if score_matmul.op != "MatMul" or context_matmul.op != "MatMul": continue

                # --- 2. 深度溯源函数：通过形状和最大搜索范围定位真正的 Input ---
                def find_true_origin(start_node, depth_limit=16):
                    curr = start_node
                    collected_in_branch = []

                    for _ in range(depth_limit):
                        if curr not in collected_in_branch:
                            collected_in_branch.append(curr)

                        # 优先识别 branching 算子
                        if curr.op in ["Gather", "Split", "Squeeze"]:
                            # 如果是 Squeeze，先看它上一级是不是 Split
                            if curr.op == "Squeeze":
                                parent_node = curr.inputs[0].inputs[0] if curr.inputs[0].inputs else None
                                if parent_node and parent_node.op == "Split":
                                    # 记录 Split 节点并继续向上
                                    if parent_node not in collected_in_branch:
                                        collected_in_branch.append(parent_node)
                                    target = parent_node.inputs[0].inputs[0]
                                    return target, collected_in_branch
                                # 如果 Squeeze 上面不是 Split，继续按普通节点追溯

                            elif curr.op == "Split":
                                # Split 的输入即源头
                                return curr.inputs[0].inputs[0], collected_in_branch

                            elif curr.op == "Gather":
                                return curr.inputs[0].inputs[0], collected_in_branch

                        # 普通追溯
                        if not curr.inputs or not isinstance(curr.inputs[0], gs.Variable) or not curr.inputs[0].inputs:
                            break
                        curr = curr.inputs[0].inputs[0]
                    return curr, collected_in_branch

                # 分别从 Q, K, V 的输入点向上追溯
                origin_q, branch_q = find_true_origin(score_matmul.inputs[0].inputs[0])
                origin_k, branch_k = find_true_origin(score_matmul.inputs[1].inputs[0])

                # 寻找 V 分支输入
                v_inputs = [inp.inputs[0] for inp in context_matmul.inputs if inp != softmax_node.outputs[0]]
                if not v_inputs: continue
                origin_v, branch_v = find_true_origin(v_inputs[0])

                # --- 3. 强校验：Q, K, V 必须汇聚于同一个源头 ---
                if not (origin_q == origin_k == origin_v):
                    # 如果不相等，说明有的分支还没追溯够，或者这根本不是 Self-Attention
                    continue

                common_source = origin_q

                # 汇总所有要清理的节点
                all_fused_nodes = [score_matmul, softmax_node, context_matmul]
                all_fused_nodes.extend(branch_q)
                all_fused_nodes.extend(branch_k)
                all_fused_nodes.extend(branch_v)

                # 向下找最后的输出转置（可选，有的模型没有）
                final_node = context_matmul.outputs[0].outputs[0]
                if final_node.op == "Transpose":
                    all_fused_nodes.append(final_node)
                    final_out_var = final_node.outputs[0]
                else:
                    final_out_var = context_matmul.outputs[0]

                # 从QKV源头向上寻找投影部分
                prev_node = common_source.inputs[0].inputs[0]
                found_bias, found_weights = False, False
                if prev_node.op == "Reshape":
                    all_fused_nodes.append(prev_node)
                    add_node = prev_node.inputs[0].inputs[0]
                    next_var = [var for var in add_node.inputs if isinstance(var, gs.Variable)][0]
                    matmul_node = next_var.inputs[0]
                else:
                    if prev_node.op != "Add":
                        continue
                    add_node = prev_node
                    matmul_node = add_node.inputs[0].inputs[0]
                if add_node.op == "Add":
                    found_bias = True
                    all_fused_nodes.append(add_node)
                if matmul_node.op == "MatMul":
                    found_weights = True
                    all_fused_nodes.append(matmul_node)
                if not found_bias or not found_weights:
                    continue

                # --- 4. 执行融合手术 ---
                # 统一获取 num_heads (从 common_source 那个汇合转置节点提取)
                num_heads = 1
                if common_source.op == "Transpose" and "perm" in common_source.attrs:
                    if len(common_source.outputs[0].shape) >= 3:
                        num_heads = common_source.outputs[0].shape[2]

                input_x = matmul_node.inputs[0] # Reshape 之前的原始数据
                attention_weight = [var for var in matmul_node.inputs if isinstance(var, gs.Constant)][0]
                attention_bias = [var for var in add_node.inputs if isinstance(var, gs.Constant)][0]

                new_attn_node = gs.Node(
                    op="Attention",
                    name=f"Fused_Attention_at_{softmax_node.name}",
                    inputs=[input_x, attention_weight, attention_bias],
                    outputs=[gs.Variable(name=f"{final_out_var.name}_fused")],
                    attrs={"num_heads": int(num_heads)},
                    domain="convdog.ai"
                )
                graph.nodes.append(new_attn_node)

                # 下游接管
                for next_node in list(final_out_var.outputs):
                    for i, inp in enumerate(next_node.inputs):
                        if inp == final_out_var:
                            next_node.inputs[i] = new_attn_node.outputs[0]

                # --- 5. 日志与清理 ---
                # 去重日志节点名
                unique_names = []
                unique_nodes = []
                for n in all_fused_nodes:
                    if n.name not in unique_names:
                        unique_names.append(n.name)
                        unique_nodes.append(n)

                logger.debug(f"[O2] 成功识别并融合 Attention 算子链: " +
                             f"{' + '.join(sorted([f'{n.name}({n.op})' for n in unique_nodes]))}")

                for n in unique_nodes:
                    n.inputs.clear()
                    n.outputs.clear()

                changed = True

            except (IndexError, AttributeError, TypeError):
                continue

        if changed:
            self.replace_custom_ops(graph)
            graph.cleanup().toposort()

        return changed

    def replace_custom_ops(self, graph: gs.Graph) -> None:
        """ 保持原有的 Custom_Op_Map 转换逻辑 """
        for node in graph.nodes:
            if node.op in self.Custom_Op_Map:
                node_op = self.Custom_Op_Map[node.op]
                domain, op = node_op.split(":")
                node.op = op
                node.domain = domain
                node.version = 1



import numpy as np
import onnx_graphsurgeon as gs

from convdog.simplifier.base_pass import BasePass
from convdog.utils.logger import logger


class FuseConvPass(BasePass):
    def process_conv1_convk(
            self, conv1: gs.Node,
            consumer: gs.Node
    ) -> None:
        """
        处理 Conv(1x1, S=1) -> Conv(KxK, S=S2) 的融合。
        融合后的结果是一个 KxK 卷积，继承第二层的 Stride 和 Padding。
        """
        # --- 1. 提取权重和偏置 ---
        # w1: (CO1, CI1, 1, 1)
        # w2: (CO2, CO1, K, K)
        w1 = conv1.inputs[1].values
        w2 = consumer.inputs[1].values

        # 提取偏置，若无则初始化为 0
        b1 = conv1.inputs[2].values if len(conv1.inputs) > 2 else np.zeros(w1.shape[0], dtype=w1.dtype)
        b2 = consumer.inputs[2].values if len(consumer.inputs) > 2 else np.zeros(w2.shape[0], dtype=w2.dtype)

        # --- 2. 核心数学融合逻辑 ---

        # A. 权重融合 (Weight Fusion)
        # 公式: W_new[o, i, h, w] = sum_k( W2[o, k, h, w] * W1[k, i, 0, 0] )
        # 将 w1 重塑为 (CO1, CI1) 以便进行矩阵乘法/einsum
        w1_linear = w1.reshape(w1.shape[0], w1.shape[1])
        # 使用 einsum 实现跨通道合并：o=CO2, k=CO1, i=CI1, h=K, w=K
        w_new = np.einsum('okhw, ki -> oihw', w2, w1_linear)

        # B. 偏置融合 (Bias Fusion)
        # 公式: b_new = Conv2(b1) + b2.
        # 由于 b1 在空间上是常数，等价于: b_new = (W2在空间维度求和) @ b1 + b2
        w2_spatial_sum = np.sum(w2, axis=(2, 3)) # 形状: (CO2, CO1)
        b_new = (w2_spatial_sum @ b1) + b2

        # --- 3. 属性继承 ---
        # 1x1 卷积不改变感受野和空间尺寸，因此融合后的算子完全继承第二层的空间属性
        new_pads = consumer.attrs.get("pads", [0, 0, 0, 0])
        new_strides = consumer.attrs.get("strides", [1, 1])
        new_dilations = consumer.attrs.get("dilations", [1, 1])

        # --- 4. 更新图结构 ---
        # 更新权重
        conv1.inputs[1].values = w_new

        # 更新或注入偏置
        if len(conv1.inputs) > 2:
            conv1.inputs[2].values = b_new
        else:
            bias_const = gs.Constant(name=f"{conv1.name}_fused_bias", values=b_new)
            conv1.inputs.append(bias_const)

        # 更新 Conv1 的属性使之接管 Conv2 的职能
        conv1.attrs["pads"] = new_pads
        conv1.attrs["strides"] = new_strides
        conv1.attrs["dilations"] = new_dilations
        # kernel_shape 会由下游框架根据权重 shape 自动推导，或手动指定
        if "kernel_shape" in consumer.attrs:
            conv1.attrs["kernel_shape"] = consumer.attrs["kernel_shape"]

        # --- 5. 拓扑接管 (Wiring) ---

        # 获取 Conv2 的输出变量
        conv2_out_var = consumer.outputs[0]
        conv1_out_var = conv1.outputs[0]

        # 同步 shape/dtype 信息
        conv1_out_var.shape = conv2_out_var.shape
        conv1_out_var.dtype = conv2_out_var.dtype

        # 将所有原本消费 Conv2 的节点，重定向至消费 Conv1
        for next_node in list(conv2_out_var.outputs):
            for i, node_input in enumerate(next_node.inputs):
                if node_input == conv2_out_var:
                    next_node.inputs[i] = conv1_out_var

        # 断开 Conv2 节点的连接，以便 graph.cleanup() 彻底移除它
        consumer.inputs = []
        consumer.outputs = []

        logger.debug(f"[O2] 成功融合Conv(1x1->KxK): {conv1.name} + {consumer.name} -> {conv1.name}")

    def process_conv1_conv1(
            self, conv1: gs.Node,
            consumer: gs.Node
    ) -> None:
        # --- 1. 提取权重与形状 ---
        w1 = conv1.inputs[1].values  # (CO1, CI1, 1, 1)
        w2 = consumer.inputs[1].values # (CO2, CO1, 1, 1)

        # 使用 reshape 替代 squeeze，安全转成 (Out, In) 的 2D 矩阵
        co2, co1_w2 = w2.shape[0], w2.shape[1]
        co1_w1, ci1 = w1.shape[0], w1.shape[1]

        w1_mat = w1.reshape(co1_w1, ci1)
        w2_mat = w2.reshape(co2, co1_w2)

        # --- 2. 权重融合 (Matrix Multiplication) ---
        # (CO2, CO1) @ (CO1, CI1) -> (CO2, CI1)
        w_new_mat = w2_mat @ w1_mat
        # 重新打回 4D Tensor 形状: (CO2, CI1, 1, 1)
        w_new = w_new_mat.reshape(co2, ci1, 1, 1)

        # --- 3. 偏置融合 ---
        b1 = conv1.inputs[2].values if len(conv1.inputs) > 2 else np.zeros(co1_w1, dtype=w1.dtype)
        b2 = consumer.inputs[2].values if len(consumer.inputs) > 2 else np.zeros(co2, dtype=w2.dtype)

        # b_new = W2 * b1 + b2
        b_new = (w2_mat @ b1) + b2

        # --- 4. 属性继承 ---
        # 1x1 -> 1x1 融合，Padding 恒定为 0 (ONNX 默认为 [0,0,0,0])
        new_pads = [0, 0, 0, 0]
        # 步长继承自第二层
        new_strides = consumer.attrs.get("strides", [1, 1])
        new_dilations = consumer.attrs.get("dilations", [1, 1])

        # --- 5. 图结构更新 ---
        conv1.inputs[1].values = w_new
        if len(conv1.inputs) > 2:
            conv1.inputs[2].values = b_new
        else:
            # 注入新常量作为 Bias
            bias_const = gs.Constant(name=f"{conv1.name}_fused_bias", values=b_new)
            conv1.inputs.append(bias_const)

        # 更新 Conv1 的关键属性
        conv1.attrs["pads"] = new_pads
        conv1.attrs["strides"] = new_strides
        conv1.attrs["dilations"] = new_dilations
        conv1.attrs["kernel_shape"] = [1, 1]

        # --- 6. 下游节点接管 ---
        conv2_out = consumer.outputs[0]
        conv1_out = conv1.outputs[0]

        # 同步元数据
        conv1_out.shape = conv2_out.shape
        conv1_out.dtype = conv2_out.dtype

        # 将 Conv2 的消费者全部指向 Conv1
        for next_node in list(conv2_out.outputs):
            for i, inp in enumerate(next_node.inputs):
                if inp == conv2_out:
                    next_node.inputs[i] = conv1_out

        # 清除 Conv2（consumer）的连接，等待 cleanup
        consumer.inputs, consumer.outputs = [], []

        logger.debug(f"[O2] 成功融合Conv(1x1->1x1): {conv1.name} + {consumer.name} -> {conv1.name}")

    def process_conv_conv(self, graph: gs.Graph) -> bool:
        changed = False
        for node in graph.nodes:
            if node.op != "Conv": continue
            if len(node.outputs) != 1 or len(node.outputs[0].outputs) != 1: continue

            conv1 = node
            consumer = conv1.outputs[0].outputs[0]
            if consumer.op != "Conv": continue

            # --- 约束检查 ---
            k1 = conv1.attrs.get("kernel_shape", [1, 1])
            k2 = consumer.attrs.get("kernel_shape", [1, 1])
            s1 = conv1.attrs.get("strides", [1, 1])
            s2 = consumer.attrs.get("strides", [1, 1])
            # O2 只处理常规卷积
            cur_size0 = k1[0]
            for k in k1:
                if k != cur_size0:
                    continue

            cur_size1 = k2[0]
            for k in k2:
                if k != cur_size1:
                    continue

            cur_stride0 = s1[0]
            for s in s1:
                if s != cur_stride0:
                    continue

            cur_stride1 = s2[0]
            for s in s2:
                if s != cur_stride1:
                    continue

            if not (isinstance(conv1.inputs[1], gs.Constant) and isinstance(consumer.inputs[1], gs.Constant)):
                continue

            # --- 判断情况 ---
            if cur_size0 == 1 and cur_size1 == 1 and cur_stride0 == 1:
                self.process_conv1_conv1(
                    conv1,
                    consumer
                )
            if cur_size0 == 1 and cur_size1 > 1 and cur_stride0 == 1:
                self.process_conv1_convk(
                    conv1,
                    consumer
                )

            graph.cleanup().toposort()
            changed = True

        return changed

    def process(self, graph: "gs.Graph") -> bool:
        """
        融合卷积层。
        """

        graph.cleanup().toposort()
        logger.debug("[O2]: 正在融合Conv+Conv......")
        changed = self.process_conv_conv(graph)

        return changed



from typing import Optional
from convdog.core.graph import ConvDogModel
from convdog.utils.logger import logger
from convdog.simplifier.fuse_consecutive_node_pass import FuseConsecutiveNodePass
from convdog.simplifier.dead_code_elimination_pass import DeadCodeEliminationPass
from convdog.simplifier.inverse_op_node_pass import InverseOpCancellationPass
from convdog.simplifier.fuse_multiple_gather_pass import FuseMultipleGatherPass
from convdog.simplifier.fuse_gemm_pass import GemmFusionPass
from convdog.simplifier.base_pass import BasePass


class O1Optimizer(object):
    def __init__(self, model: ConvDogModel, safe_mode=False):
        self.model = model
        self.safe_mode = safe_mode
        self.fuse_consecutive_node: Optional[BasePass] = None
        self.dead_code_elimination_pass: Optional[DeadCodeEliminationPass] = None
        self.gemm_pass: Optional[GemmFusionPass] = None
        self.multi_gather_pass: Optional[FuseMultipleGatherPass] = None
        self.inverse_pass: Optional[InverseOpCancellationPass] = None
        self.initialize_pass()

    def initialize_pass(self):
        self.fuse_consecutive_node = FuseConsecutiveNodePass()
        self.dead_code_elimination_pass = DeadCodeEliminationPass()
        self.gemm_pass = GemmFusionPass()
        self.multi_gather_pass = FuseMultipleGatherPass()
        self.inverse_pass = InverseOpCancellationPass()

    def apply(self) -> ConvDogModel:
        logger.info("[O1]: 开始执行初步算子优化和拓扑结构优化...")

        # 记录优化前的初始节点数
        prev_node_count = len(self.model.graph.nodes)
        iteration = 0

        while True:
            iteration += 1

            # --- 执行优化 Pass ---
            self.fuse_consecutive_node.run(self.model)
            self.dead_code_elimination_pass.run(self.model)
            self.gemm_pass.run(self.model)
            self.inverse_pass.run(self.model)
            # self.multi_gather_pass.run(self.model)

            self.model.reset_value_info()
            self.model.fold_tensors()

            # 获取优化后的节点数量
            current_node_count = len(self.model.graph.nodes)
            diff = prev_node_count - current_node_count
            diff_color = "green" if diff > 0 else "white"
            logger.info(
                f"[第 {iteration} 轮优化]："
                f"节点数 {prev_node_count} -> {current_node_count} "
                f"[{diff_color}]已削减:[/] {diff}"
            )

            # [收敛收工判断]
            # 如果节点数量不再发生变化，说明图中已经没有连续的节点了
            if current_node_count == prev_node_count:
                logger.success(f"[O1]: 优化收敛！总共迭代 {iteration} 轮。")
                break

            # 更新计数，继续下一轮嗅探
            prev_node_count = current_node_count

            # 防止死循环（保险起见，设置一个最大迭代次数）
            if iteration > 100:
                logger.warning("[O1]: 迭代次数过多，强制跳出，请检查图中是否存在环路。")
                break

        return self.model



import numpy as np
import onnx_graphsurgeon as gs

from convdog.simplifier.base_pass import BasePass
from convdog.utils.logger import logger


class FuseGeluPass(BasePass):
    """
    针对常见深度学习框架导出的 GELU 模式进行融合。
    匹配模式: x -> Div(sqrt(2)) -> Erf -> Add(1) -> Mul(x) -> Mul(0.5) -> Gelu(x)
    """
    Custom_Op_Map = {
        "Gelu": "com.microsoft:Gelu"
    }

    def _is_constant_val(self, input_tensor, target, tolerance=1e-3):
        """ 辅助函数：检查张量是否为特定常数 """
        if isinstance(input_tensor, gs.Constant):
            val = input_tensor.values
            return np.all(np.abs(val - target) < tolerance)
        return False

    def process_gelu_fusion(self, graph: gs.Graph) -> bool:
        changed = False
        # 为了避免遍历时修改导致的迭代器问题，先固定节点列表
        # 我们以 Erf 算子作为锚点（Anchor），因为它在 Gelu 结构中特征最明显
        for node in [n for n in graph.nodes if n.op == "Erf"]:
            # --- 结构匹配逻辑 ---
            # 1. 回溯 Div: Erf <- Div
            div_node = node.inputs[0].inputs[0] if node.inputs[0].inputs else None
            if not div_node or div_node.op != "Div":
                continue

            # 校验 Div 的常数是否接近 sqrt(2) (约 1.414)
            if not any(self._is_constant_val(inp, 1.4142, 1e-2) for inp in div_node.inputs):
                continue

            # 2. 向下寻找 Add: Erf -> Add
            if len(node.outputs[0].outputs) != 1: continue
            add_node = node.outputs[0].outputs[0]
            if add_node.op != "Add" or not any(self._is_constant_val(inp, 1.0) for inp in add_node.inputs):
                continue

            # 3. 向下寻找第一个 Mul (合并分支): Add -> Mul_0
            if len(add_node.outputs[0].outputs) != 1: continue
            mul0_node = add_node.outputs[0].outputs[0]
            if mul0_node.op != "Mul": continue

            # 验证 Mul0 的另一个输入是否是最初的 x (即 Div 的输入)
            input_x = div_node.inputs[0]
            if input_x not in mul0_node.inputs:
                continue

            # 4. 向下寻找最后一个 Mul (缩放 0.5): Mul0 -> Mul_final
            if len(mul0_node.outputs[0].outputs) != 1: continue
            mul_final_node = mul0_node.outputs[0].outputs[0]
            if mul_final_node.op != "Mul" or not any(self._is_constant_val(inp, 0.5) for inp in mul_final_node.inputs):
                continue

            # --- 命中优化条件，开始执行手术 ---
            # 创建新的 Gelu 节点
            # 使用 Microsoft 的 Gelu 定义，输入是最初的 x
            gelu_out_var = gs.Variable(name=f"{mul_final_node.name}_fused_gelu_out")
            new_gelu_node = gs.Node(
                op="Gelu",
                name=f"{node.name}_fused_gelu",
                inputs=[input_x],
                outputs=[gelu_out_var],
                domain="convdog.ai" # 标记为自定义算子
            )
            graph.nodes.append(new_gelu_node)

            # 接管输出
            final_out_var = mul_final_node.outputs[0]
            for next_node in list(final_out_var.outputs):
                for i, inp in enumerate(next_node.inputs):
                    if inp == final_out_var:
                        next_node.inputs[i] = gelu_out_var

            # 清理旧节点连接
            for n in [div_node, node, add_node, mul0_node, mul_final_node]:
                n.inputs, n.outputs = [], []

            logger.debug(
                f"[O2] 成功发现并融合标准Gelu结构: "
                f"{input_x.name} + {div_node.name} + "
                f"{node.name} + {add_node.name} + "
                f"{mul0_node.name} + {mul_final_node.name} "
                f"-> {new_gelu_node.name}"
            )
            changed = True

        return changed

    def replace_custom_ops(self, graph: gs.Graph) -> None:
        """
        统一将内部标记的 Gelu 节点转换为目标后端支持的 domain
        """
        for node in graph.nodes:
            if node.op in self.Custom_Op_Map:
                target = self.Custom_Op_Map[node.op]
                domain, op = target.split(":")
                node.op = op
                node.domain = domain

    def process(self, graph: "gs.Graph") -> bool:
        """
        融合Gelu层。
        """
        graph.cleanup().toposort()

        # 1. 执行模式匹配与融合
        changed = self.process_gelu_fusion(graph)

        # 2. 如果发生了变化，执行替换和清理
        if changed:
            self.replace_custom_ops(graph)
            graph.cleanup().toposort()

        return changed



import argparse
import sys
import time
from typing import Dict, Optional

from convdog.core.graph import ConvDogModel
from convdog.optimizer.O0 import O0Optimizer
from convdog.optimizer.O1 import O1Optimizer
from convdog.optimizer.O2 import O2Optimizer
from convdog.optimizer.O3 import O3Optimizer
from convdog.utils.logger import logger
from convdog.utils.stats import (
    ModelStats,
    print_comparison_table,
    print_quant_summary
)
from convdog.core.typing_extension import BackendType


def parse_shape_arg(shape_str) -> Optional[Dict[str, int]]:
    """解析 image:1,3,224,224 格式"""
    if not shape_str: return None
    try:
        res = {}
        for item in shape_str.split(";"):
            name, dims = item.split(":")
            res[name] = [int(d) for d in dims.split(",")]
        return res
    except Exception as e:
        logger.error(e)
        raise argparse.ArgumentTypeError("形状格式必须为 'name:1,3,224,224'")


def parse_level_arg(level_str) -> int:
    """解析优化等级"""
    if level_str not in ["O0", "O1", "O2", "O3"]:
        logger.error("[x] 暂不支持的优化等级!!!")
        sys.exit(-1)
    return {"O0": 0, "O1": 1, "O2": 2, "O3": 3}[level_str]


def parse_backend_arg(backend_str) -> BackendType:
    if backend_str == "default":
        return BackendType.DEFAULT
    elif backend_str == "qnn":
        return BackendType.QNN
    else:
        logger.error("[x] 暂不支持的O3级别后端优化, 请尝试O2等级优化!")
        sys.exit(-1)


def optimize_model(
        input_path: str,
        output_path: str,
        opt_level: int = 0,
        input_shapes: Optional[Dict[str, int]] = None,
        fp16: bool = True,
        backend: BackendType = BackendType.DEFAULT
):
    # 加载原始模型
    graph = ConvDogModel(input_path)
    original_stats = ModelStats(graph, input_path)
    o0_optimizer = O0Optimizer(graph, input_shapes)
    o1_optimizer, o2_optimizer, o3_optimizer = None, None, None

    # 注入信息
    graph.inject_convdog_info()

    start_time = time.time()
    optimized_graph = o0_optimizer.apply()
    logger.success(f"[*] O0等级优化完毕!")

    if opt_level >= 1:
        o1_optimizer = O1Optimizer(optimized_graph)
        optimized_graph = o1_optimizer.apply()
        logger.success(f"[*] O1等级优化完毕!")
    if opt_level >= 2:
        o2_optimizer = O2Optimizer(optimized_graph)
        optimized_graph = o2_optimizer.apply()
        logger.success(f"[*] O2等级优化完毕!")

    if opt_level >= 3:
        o3_optimizer = O3Optimizer(optimized_graph, backend)
        optimized_graph = o3_optimizer.apply()
        logger.success(f"[*] O3等级优化完毕!")

    if opt_level >= 2:
        o2_optimizer.replace_custom_ops()
        graph.sync_model()

    if opt_level >= 3:
        o3_optimizer.replace_custom_ops()
        graph.sync_model()

    elapsed = time.time() - start_time

    # 保存优化后模型
    optimized_stats = ModelStats(graph, output_path)
    logger.info("正在统计计算图优化情况......")
    print_comparison_table(original_stats, optimized_stats, elapsed)
    logger.info("正在统计推理指标......")
    print_quant_summary(original_stats, optimized_stats, input_path)
    optimized_graph.save(output_path)


def main():
    parser = argparse.ArgumentParser(description="ConvDog🐕 模型优化工具")
    parser.add_argument("input", help="输入 ONNX 路径")
    parser.add_argument("output", help="输出 ONNX 路径")
    parser.add_argument("level", type=parse_level_arg, help="优化等级, 可填O0~O3")
    parser.add_argument("--shapes", type=parse_shape_arg, help="静态化形状, 格式 'name:1,3,224,224'")
    parser.add_argument("--no_fp16", action="store_true", help="fp16量化, 默认在O2阶段开启")
    parser.add_argument(
        "--backend", default=BackendType.DEFAULT,
        type=parse_backend_arg, help="O3阶段选择的目标优化后端"
    )
    args = parser.parse_args()
    fp16 = not args.no_fp16

    optimize_model(
        args.input,
        args.output,
        args.level,
        args.shapes,
        fp16,
        args.backend
    )


if __name__ == '__main__':
    sys.exit(main())



from convdog.simplifier.base_pass import BasePass
from convdog.utils.logger import logger
import onnx_graphsurgeon as gs

class GemmFusionPass(BasePass):
    """
    O1 级优化：将 MatMul + Add 融合为单个 Gemm 节点。
    公式：Y = alpha * A * B + beta * C
    """
    def process(self, graph: "gs.Graph") -> bool:
        changed = False
        # 在开始前进行清理，确保节点的 outputs (consumers) 索引是最新的
        graph.cleanup().toposort()

        # 1. 遍历图中所有节点，寻找 MatMul
        # 使用列表推导式创建快照，防止遍历过程中修改图导致迭代器失效
        for node in [n for n in graph.nodes if n.op == "MatMul"]:

            # --- [关键修复：秩检查] ---
            # 获取 MatMul 的两个输入 A 和 B
            A = node.inputs[0]
            B = node.inputs[1]

            # 核心逻辑：Gemm 只支持 2D 矩阵乘法。
            # Transformer 中的 MatMul 往往是 3D [Batch, Seq, Dim]，这种不能融合为 Gemm
            # 如果 shape 为 None (推断失败) 或 长度不等于 2，则跳过
            if A.shape is None or len(A.shape) != 2:
                continue
            if B.shape is None or len(B.shape) != 2:
                continue
            # ------------------------

            # 2. 检查 MatMul 的输出是否只连接到了一个节点
            if len(node.outputs) != 1:
                continue

            # 在 gs 中，Variable.outputs 存放的就是该张量的消费者节点 (consumers)
            consumers = node.outputs[0].outputs
            if len(consumers) != 1:
                continue

            consumer = consumers[0]

            # 3. 检查下游节点是否为 Add
            if consumer.op == "Add":
                # --- 进阶优化：检查 B 是否来自一个 Transpose 节点 ---
                trans_b = 0
                # 如果 B 的产生者是一个 Transpose 节点，且属性是 [1, 0]
                if len(B.inputs) == 1 and B.inputs[0].op == "Transpose":
                    trans_node = B.inputs[0]
                    if trans_node.attrs.get("perm") == [1, 0]:
                        # 直接吃掉 Transpose，将数据源上移，并在 Gemm 属性中标记 transB=1
                        B = trans_node.inputs[0]
                        trans_b = 1
                        logger.debug(f"[O1] 成功吃掉权重的 Transpose: {trans_node.name}")

                # 4. 寻找 Add node 中非 MatMul 输出的那个输入作为 Bias (C)
                C = None
                for inp in consumer.inputs:
                    if inp.name != node.outputs[0].name:
                        # 增加判断：Bias 应该是合法的常量或 Initializer
                        if isinstance(inp, gs.Constant) or (hasattr(inp, 'values')):
                            C = inp
                            break

                if C is None:
                    logger.debug(f"[O1] 跳过融合：找不到合法的 Bias 输入 {node.name}")
                    continue

                # 5. 创建新的 Gemm 节点
                gemm_node = gs.Node(
                    op="Gemm",
                    name=f"Fused_Gemm_{node.name}",
                    attrs={"alpha": 1.0, "beta": 1.0, "transA": 0, "transB": trans_b},
                    inputs=[A, B, C],
                    outputs=consumer.outputs
                )

                # 6. 将新节点加入图，并排干旧节点的引用
                graph.nodes.append(gemm_node)

                # 这种“清空输出”的操作能让接下来的 cleanup() 自动剔除孤立的旧节点和中间张量
                node.outputs = []
                node.inputs = []
                consumer.outputs = []
                consumer.inputs = []

                changed = True
                logger.debug(f"[O1] 成功融合: {node.name} + {consumer.name} -> Gemm")

        # 7. 手术结束后再次清理图
        if changed:
            graph.cleanup().toposort()

        return changed



import onnx_graphsurgeon as gs
from convdog.simplifier.base_pass import BasePass
from convdog.utils.logger import logger

class FuseConsecutiveNodePass(BasePass):
    """
    O1 Pass: 融合连续的恒等算子。
    举例：如果一个 ReLU 的输入来自于另一个 ReLU 的输出，则跳过第二个 ReLU。
    """

    IDEMPOTENT_OPS = {
        "Relu": [],
        "LeakyRelu": ["alpha"],
        "Clip": ["min", "max"],
        "Abs": [],
        "Sign": [],
        "Floor": [],
        "Ceil": [],
        "Identity": [],
        "Dropout": []
    }

    def process_identity(self, graph: gs.Graph) -> bool:
        # 只处理f(f(x)) = f(x)的变换

        # 记录融合数量
        changed = False
        fusion_count = 0

        # 遍历图中所有节点
        for node in graph.nodes:
            if node.op in self.IDEMPOTENT_OPS:
                # 获取该节点的输入 Tensor
                input_tensor = node.inputs[0]
                cur_op = node.op

                # 检查产生该输入 Tensor 的上游节点是谁
                # input_tensor.inputs 返回的是产生该 tensor 的节点列表
                if len(input_tensor.inputs) > 0:
                    prev_node = input_tensor.inputs[0]

                    # 如果上游节点也是 ReLU
                    if prev_node.op == cur_op:
                        # 【核心手术】：将当前 ReLU 的所有输出，直接挂到上游 ReLU 的输入上
                        # 也就是：Relu1 -> TensorX -> Relu2 -> TensorY
                        # 变成：Relu1 -> TensorY

                        # 检查关键属性是否一致
                        attrs = self.IDEMPOTENT_OPS[node.op]
                        for attr in attrs:
                            if node.attrs.get(attr) != prev_node.attrs.get(attr):
                                break

                        # 把当前节点的所有后继消费者的输入，指向前一个节点的输出
                        # 在 graphsurgeon 中，最简单的方法是合并输出 Tensor
                        prev_node.outputs = node.outputs

                        # 清空当前节点的输出，方便后面 cleanup 自动回收
                        node.outputs = []
                        node.inputs = []
                        fusion_count += 1
                        logger.debug(f"[O1] 成功融合: {node.name} + {prev_node.name} -> {prev_node.name}")
                        changed = True

        if fusion_count > 0:
            # cleanup 会自动清除因为失去输出连接而变成死节点的“第二个节点”
            graph.cleanup()
            logger.debug(f"[O1]: 成功切除了 {fusion_count} 个冗余的恒等节点！")
        else:
            logger.debug("[O1]: 未发现连续的恒等节点，保持原样。")

        return changed

    def process_gemm(self, graph: gs.Graph) -> bool:
        fusion_count = 0
        changed = False
        for node in graph.nodes:
            if node.op == "Gemm":
                # 检查是否只连接一个算子
                if len(node.outputs) != 1:
                    continue

                # 查看是否为单消费者
                consumers = node.outputs[0].outputs
                if len(consumers) != 1:
                    continue

                consumer = consumers[0]
                if consumer.op != "Gemm":
                    continue

                can_fuse = isinstance(node.inputs[1], gs.Constant) and \
                           isinstance(consumer.inputs[1], gs.Constant)
                if len(consumer.inputs) == 3:
                    can_fuse = can_fuse and isinstance(consumer.inputs[2], gs.Constant)
                if len(node.inputs) == 3:
                    can_fuse = can_fuse and isinstance(node.inputs[2], gs.Constant)
                if can_fuse:
                    w1 = node.inputs[1].values
                    w2 = consumer.inputs[1].values
                    b1 = node.inputs[2].values if len(node.inputs) > 2 else 0
                    b2 = consumer.inputs[2].values if len(consumer.inputs) > 2 else 0
                    if node.attrs.get("transB", 0): w1 = w1.T
                    if consumer.attrs.get("transB", 0): w2 = w2.T
                    alpha1, beta1 = node.attrs.get("alpha", 1.0), node.attrs.get("beta", 1.0)
                    alpha2, beta2 = consumer.attrs.get("alpha", 1.0), consumer.attrs.get("beta", 1.0)

                    alpha = alpha1 * alpha2
                    beta = 1.0
                    a = node.inputs[0]
                    b = w1 @ w2
                    c = alpha2 * beta1 * (b1 @ w2) + beta2 * b2
                    trans_a = node.attrs.get("transB", 0)
                    trans_b = 0

                    # 创建融合后节点
                    b = gs.Constant(
                        "w1",
                        b
                    )
                    c = gs.Constant(
                        "b1",
                        c
                    )
                    gemm_node = gs.Node(
                        op="Gemm",
                        name=f"Fused_Gemm_{node.name}",
                        attrs={"alpha": alpha, "beta": beta, "transA": trans_a, "transB": trans_b},
                        inputs=[a, b, c],
                        outputs=consumer.outputs
                    )

                    graph.nodes.append(gemm_node)
                    node.outputs = []
                    node.inputs = []
                    consumer.outputs = []
                    consumer.inputs = []
                    graph.cleanup().toposort()
                    fusion_count += 1
                    changed = True
                    logger.debug(f"[O1] 成功融合: {node.name} + {consumer.name} -> Gemm")

        return changed

    def process_non_identity(self, graph: gs.Graph) -> bool:
        changed = self.process_gemm(graph)
        return changed

    def process(self, graph: gs.Graph) -> bool:
        # 在开始前进行清理，确保节点的 outputs (consumers) 索引是最新的
        graph.cleanup().toposort()
        logger.debug("[O1]: 正在嗅探并融合连续重复的节点...")
        logger.debug("[O1]: 处理恒等变换节点......")
        changed0 = self.process_identity(graph)
        logger.debug("[O1]: 处理非恒等变换节点......")
        changed1 = self.process_non_identity(graph)
        changed = changed0 or changed1

        return changed



import onnx_graphsurgeon as gs

from convdog.simplifier.base_pass import BasePass
from convdog.utils.logger import logger


class InverseOpCancellationPass(BasePass):
    """
    O1 Pass: 清除互逆的算子。
    举例：如果检测到先squeeze(1)再unsqueeze(1)的操作，则直接清除
    """
    Inverse_Ops = {
        "Squeeze": "Unsqueeze",
        "Unsqueeze": "Squeeze",
        "Concat": "Split",
        "Split": "Concat",
    }

    def process_squeeze_unsqueeze(self, graph: gs.Graph) -> bool:
        changed = False
        fusion_count = 0

        for node in graph.nodes:
            if node.op in ["Squeeze", "Unsqueeze"]:
                op_type = node.op
                if len(node.inputs) < 2:
                    continue
                node_axis = node.inputs[1].values.item()
                # 检查是否只连接一个算子
                if len(node.outputs) != 1:
                    continue

                # 查看是否为单消费者
                consumers = node.outputs[0].outputs
                if len(consumers) != 1:
                    continue

                consumer = consumers[0]
                if consumer.op != self.Inverse_Ops[op_type]:
                    continue

                consumer_axis = consumer.inputs[1].values.item()
                can_dce = consumer_axis == node_axis

                if can_dce:
                    prev_var = node.inputs[0]
                    next_var = consumer.outputs[0]
                    for next_node in next_var.outputs:
                    # 找到这个消费者节点中，哪个输入位置是 T3，然后把它换成 T1
                        for i, node_input in enumerate(next_node.inputs):
                            if node_input == next_var:
                                next_node.inputs[i] = prev_var
                    node.inputs = []
                    node.outputs = []
                    consumer.inputs = []
                    consumer.outputs = []

                    graph.cleanup().toposort()
                    fusion_count += 1
                    changed = True
                    logger.debug(f"[O1] 成功消除: {node.name} + {consumer.name} -> None")

        return changed

    def process_split_concat(self, graph: gs.Graph) -> bool:
        changed = False
        fusion_count = 0

        # 使用 list(graph.nodes) 避免在迭代时修改图结构导致的迭代器失效
        for node in list(graph.nodes):
            if node.op == "Split":
                # 1. 获取 Split 的 Axis (通常在 attrs 里)
                split_axis = node.attrs.get("axis", 0)
                split_outputs = node.outputs

                if len(split_outputs) == 0:
                    continue

                # 2. 检查 Split 的所有输出是否都只去往同一个消费者，且该消费者是 Concat
                # 先看第一个输出的第一个消费者
                first_out_consumers = split_outputs[0].outputs
                if len(first_out_consumers) != 1:
                    continue

                consumer = first_out_consumers[0]
                if consumer.op != "Concat":
                    continue

                # 3. 核心校验：
                # a) Concat 的 axis 必须与 Split 一致
                # b) Concat 的输入列表必须完全等于 Split 的输出列表（顺序和数量都要一致）
                # c) Split 的每个输出张量都不能有第二个消费者（旁路）
                concat_axis = consumer.attrs.get("axis", 0)
                if split_axis != concat_axis:
                    continue

                # 检查是否满足“原路拆解原路合并”：Concat的输入就是Split的输出
                if consumer.inputs != split_outputs:
                    continue

                # 检查是否有旁路消费者（这步非常重要，否则删了 Split 会导致其他支路断路）
                has_side_consumer = any(len(out.outputs) > 1 for out in split_outputs)
                if has_side_consumer:
                    continue

                # --- 命中优化条件：执行 A -> Split -> Concat -> B  =>  A -> B ---
                prev_var = node.inputs[0]   # Split 的输入
                next_var = consumer.outputs[0] # Concat 的输出

                # 将 Concat 之后的所有消费者，重新接到 Split 之前的张量上
                for next_node in list(next_var.outputs):
                    for i, node_input in enumerate(next_node.inputs):
                        if node_input == next_var:
                            next_node.inputs[i] = prev_var

                # 彻底断开旧节点的连接，以便 cleanup
                node.inputs = []
                node.outputs = []
                consumer.inputs = []
                consumer.outputs = []

                # 如果有必要，手动把原本 next_var 的元数据同步给 prev_var
                # 但通常 Split->Concat 不改变 shape/dtype，直接连即可

                graph.cleanup().toposort()
                fusion_count += 1
                changed = True
                logger.debug(f"[O1] 成功消除冗余链路: {node.name} (Split) + {consumer.name} (Concat) -> None")

        return changed

    def process(self, graph: "gs.Graph") -> bool:
        logger.debug("[O1]: 正在消除互逆算子......")
        graph.cleanup().toposort()
        changed0 = self.process_squeeze_unsqueeze(graph)
        changed1 = self.process_split_concat(graph)
        changed = changed0 or changed1

        return changed



from typing import TYPE_CHECKING

from convdog.simplifier.base_pass import BasePass
from convdog.utils.logger import logger
if TYPE_CHECKING:
    import onnx_graphsurgeon as gs


class DeadCodeEliminationPass(BasePass):
    def __init__(self):
        super().__init__()

    def process(self, graph: "gs.Graph") -> bool:
        """
        执行死代码消除：
        1. 移除输出未被使用的孤立节点（非 Graph Outputs）。
        2. 移除未被任何节点引用的 Initializers (权重)。
        3. 移除未被使用的子图输入/输出。
        """
        # 记录清理前的状态
        prev_node_count = len(graph.nodes)
        prev_const_count = len(graph.tensors().keys())

        # 核心：调用 graphsurgeon 的 cleanup 魔法
        # recursive=True 会递归删除。例如：如果 A->B，B被删了，
        # 且 A 也没有其他输出，那么 A 也会被连带删除。
        graph.toposort()
        graph.cleanup()

        # 检查是否有变化
        curr_node_count = len(graph.nodes)
        curr_const_count = len(graph.tensors().keys())

        changed = (curr_node_count != prev_node_count) or (curr_const_count != prev_const_count)

        if changed:
            logger.info(f"[O1] 清理完成: 节点 {prev_node_count}->{curr_node_count}, "
                        f"参数 {prev_const_count}->{curr_const_count}")

        return changed



import onnx_graphsurgeon as gs

from convdog.simplifier.base_pass import BasePass
from convdog.utils.logger import logger


class FuseConvPass(BasePass):
    Custom_Op_Map = {
        "ConvRelu": "com.microsoft:FusedConv"
    }

    def process(self, graph: gs.Graph) -> bool:
        changed = False
        fusion_count = 0

        logger.debug("[O3]: 正在融合Conv+Relu......")
        graph.cleanup().toposort()

        for node in graph.nodes:
            if node.op == "Conv":
                # 1. 检查 Conv 是否只有一个输出张量
                if len(node.outputs) != 1:
                    continue

                # 2. 检查输出张量是否只有一个消费者（单消费者检查）
                consumers = node.outputs[0].outputs
                if len(consumers) != 1:
                    continue

                consumer = consumers[0]
                # 3. 检查消费者是否为 ReLU
                if consumer.op != "Relu":
                    continue

                # --- 命中优化条件：执行算子融合 (Conv + ReLU -> Fused Conv) ---

                # 4. 仿照 ONNX Runtime，通过属性注入实现融合
                # 设置 activation 为 Relu，并将 auto_pad 显式设为 NOTSET（表示 Padding 已固化）
                node.op = "ConvRelu"
                node.domain = "convdog.ai"  # 标记注入自定义算子
                node.attrs["activation"] = "Relu"
                node.attrs["auto_pad"] = "NOTSET"

                # 5. 接管输出：
                # Conv 原本的输出：T_conv
                # ReLU 原本的输出：T_relu
                # 我们让后续所有原本消费 T_relu 的节点，全部改去消费 T_conv
                conv_out_var = node.outputs[0]
                relu_out_var = consumer.outputs[0]

                # 同步元数据信息，确保图的一致性
                conv_out_var.shape = relu_out_var.shape
                conv_out_var.dtype = relu_out_var.dtype

                # 找到 T_relu 的后续所有消费者
                for next_node in list(relu_out_var.outputs):
                    for i, node_input in enumerate(next_node.inputs):
                        if node_input == relu_out_var:
                            next_node.inputs[i] = conv_out_var

                # 6. 断开ReLU节点的连接，以便 cleanup 清理节点
                consumer.inputs = []
                consumer.outputs = []

                # 7. 刷新图结构
                graph.cleanup().toposort()
                fusion_count += 1
                changed = True
                logger.debug(f"[O2] 成功融合: {node.name} + {consumer.name} -> {node.name}")

        return changed

    def replace_custom_ops(self, graph: gs.Graph) -> None:
        """
        替换自定义算子为微软算子。
        """
        for node in graph.nodes:
            if node.op in self.Custom_Op_Map:
                node_op = self.Custom_Op_Map[node.op]
                domain, op = node_op.split(":")
                if domain == "No":
                    domain = None
                node.op = op
                node.domain = domain  # 变成微软算子域
                node.version = 1



import numpy as np
import onnx_graphsurgeon as gs

from convdog.simplifier.base_pass import BasePass
from convdog.utils.logger import logger

class FuseLayerNormPass(BasePass):
    """
    O2 级优化：按照非线性拓扑结构融合 LayerNorm。
    拓扑结构：
       [Input X] ----+-----> [ReduceMean] (RM1)
                     |            |
                     +-----> [Sub] <------- (分子计算)
                               |
                +--------------+------ concrete result (分子)
                |              |
              [Pow]            |
                |              |
          [ReduceMean] (RM2)   |
                |              |
            [Add eps]          |
                |              |
             [Sqrt]            |
                |              |
                +-----> [Div] <+------- (归一化)
                         |
                       [Mul] (Gamma)
                         |
                       [Add] (Beta)
    """

    def _is_constant_val(self, input_tensor, target, tolerance=1e-6):
        if isinstance(input_tensor, gs.Constant):
            return np.all(np.abs(input_tensor.values - target) < tolerance)
        return False

    def process(self, graph: "gs.Graph") -> bool:
        changed = False
        logger.debug("[O3]: 正在融合LayerNorm......")
        graph.cleanup().toposort()

        # 1. 以第一个 ReduceMean 为锚点尝试匹配
        for rm1_node in [n for n in graph.nodes if n.op == "ReduceMean"]:
            try:
                # --- [拓扑校验阶段] ---
                # 获取原始输入 X
                input_x = rm1_node.inputs[0]

                # 2. 寻找 Sub：必须同时连接 input_x 和 rm1_node 的输出
                sub_node = None
                for consumer in rm1_node.outputs[0].outputs:
                    if consumer.op == "Sub" and input_x in consumer.inputs:
                        sub_node = consumer
                        break
                if not sub_node: continue

                # 3. 寻找 Pow 和 Div：它们都是 sub_node 的消费者
                pow_node = None
                div_node = None
                for consumer in sub_node.outputs[0].outputs:
                    if consumer.op == "Pow": pow_node = consumer
                    if consumer.op == "Div": div_node = consumer

                if not pow_node or not div_node: continue
                if not self._is_constant_val(pow_node.inputs[1], 2.0): continue

                # 4. 沿着 Pow 向下寻找 RM2 -> Add(eps) -> Sqrt
                if len(pow_node.outputs[0].outputs) != 1: continue
                rm2_node = pow_node.outputs[0].outputs[0]
                if rm2_node.op != "ReduceMean": continue

                if len(rm2_node.outputs[0].outputs) != 1: continue
                add_eps_node = rm2_node.outputs[0].outputs[0]
                if add_eps_node.op != "Add": continue

                if len(add_eps_node.outputs[0].outputs) != 1: continue
                sqrt_node = add_eps_node.outputs[0].outputs[0]
                if sqrt_node.op != "Sqrt": continue

                # 5. 校验 Sqrt 的输出是否回到了刚才找到的那个 Div
                if div_node not in sqrt_node.outputs[0].outputs: continue

                # 6. 继续寻找最后的 Mul 和 Add
                if len(div_node.outputs[0].outputs) != 1: continue
                mul_node = div_node.outputs[0].outputs[0]
                if mul_node.op != "Mul": continue

                if len(mul_node.outputs[0].outputs) != 1: continue
                add_beta_node = mul_node.outputs[0].outputs[0]
                if add_beta_node.op != "Add": continue

                # --- [参数提取阶段] ---
                eps_val = float(add_eps_node.inputs[1].values.item()) if isinstance(add_eps_node.inputs[1], gs.Constant) else 1e-5
                gamma = mul_node.inputs[1]
                beta = add_beta_node.inputs[1]

                # --- [熔炼阶段] ---
                logger.debug(f"[O2] 发现非线性 LayerNorm 结构，开始融合...")

                ln_node = gs.Node(
                    op="LayerNormalization",
                    name=f"Fused_LN_at_{rm1_node.name}",
                    inputs=[input_x, gamma, beta],
                    outputs=[add_beta_node.outputs[0]],
                    attrs={"epsilon": eps_val, "axis": -1}
                )
                graph.nodes.append(ln_node)

                # --- [日志与清理] ---
                logger.debug(
                    f"[O2] 成功融合算子链: "
                    f"{rm1_node.name} + {sub_node.name} + {pow_node.name} + {rm2_node.name} + "
                    f"{add_eps_node.name} + {sqrt_node.name} + {div_node.name} + "
                    f"{mul_node.name} + {add_beta_node.name} -> {ln_node.name}"
                )

                # 断开所有参与融合节点的连接，以便 cleanup 清理
                for n in [
                    rm1_node, sub_node, pow_node,
                    rm2_node, add_eps_node, sqrt_node,
                    div_node, mul_node, add_beta_node
                ]:
                    n.outputs = []
                    n.inputs = []

                changed = True

            except (IndexError, AttributeError):
                continue

        if changed:
            graph.opset = 17 if graph.opset < 17 else graph.opset
            for entry in graph.import_domains:
                if entry.domain == "" or entry.domain == "ai.onnx":
                    if entry.version < 17:
                        logger.warning(f"[O2] 检测到 Opset {entry.version} 不支持 LayerNorm，已自动强升至 17")
                        entry.version = 17
            graph.cleanup().toposort()

        return changed
